from asyncio import (
    AbstractEventLoop,
    CancelledError,
    Event,
    Future,
    get_event_loop,
    sleep,
)
from contextlib import suppress
from dataclasses import field
from functools import partial, wraps
from random import randint
from ssl import SSLContext
from time import time
from typing import Any, ClassVar

import aiormq.exceptions as rmq
import pamqp.commands as cmd
import pamqp.exceptions as pam
from aio_pika import Connection as DumbConnection
from aio_pika import DeliveryMode, Message
from aio_pika.abc import ExchangeType
from aio_pika.connection import make_url
from aio_pika.tools import CallbackCollection
from aiormq.abc import DeliveredMessage
from fair_async_rlock import FairAsyncRLock
from kalib import (
    HTTP,
    Is,
    Logging,
    Property,
    Str,
    Who,
    dataclass,
    dumps,
    json,
    loads,
    unique,
)
from xxhash import xxh32_hexdigest
from xxhash import xxh128_hexdigest as digest

import karq.exceptions as exc
from karq.retry import TimeoutExceptions, retry
from karq.utils import attach, log_long_wait

__all__ = ('Rabbit', 'Routine', 'Timeouts')


class Tag(dataclass.mutable):

    _tag: str | None = None

    @classmethod
    def new(cls, *args, **kw):
        return cls.load(kw, *args)

    def __bool__(self):
        return bool(self._tag)

    def reset(self):
        self._tag = None

    #

    @property
    def tag(self):
        if not self._tag:
            self._tag = self.random
        return self._tag

    @property
    def random(self):
        return f'{Who(self)}.tag.{randint(0, 10000000)}'


class AsyncLock(FairAsyncRLock, Logging.Mixin):

    def __str__(self):
        msg = Who(self, addr=True)
        if self._count:
            msg = f'{msg}; depth={self._count:d}, owned by: {self._owner!r}'
        return f'<{msg}>'

    async def acquire(self):
        start = time()
        result = await super().acquire()
        spent = time() - start
        if spent > 1.0:
            self.log.debug(
                f'acquired, but too late ({spent:0.4f} sec)', trace=True)
        return result


class Timeouts(dataclass.flex):

    min_quant : float = 0.25
    min_delay : float = 5.0
    max_delay : float = 300.0
    reconnect : float = 600.0


class Connection(Logging.Mixin, DumbConnection):
    _options     : ClassVar[dict] = {}
    _connections : ClassVar[dict] = {}

    @Property.Class.Parent
    def lock(cls):
        return AsyncLock()

    @classmethod
    async def make(cls, uri, **options):
        url = make_url(
                host     = uri.host,
                port     = uri.port,
                login    = uri.user,
                password = uri.password,
                virtualhost = (uri.path.strip('/')) or '/',
                **options)

        key = f'{url}:{xxh32_hexdigest(json.dumps(options, bytes=True))}'

        async with cls.lock:
            try:
                connection = cls._connections[key]
                connection_options = dict(cls._options[key])

                if options != connection_options:
                    cls.log.warning(
                        f"couldn't reinitialize connection "
                        f'{Who(connection, addr=True)} '
                        f'({url}, {connection_options=}), with another {options=}.')

            except KeyError:
                connection = cls._connections[key] = cls(url, **options)
                cls._options[key] = options.copy()

        return connection

    async def close(self):
        async with self.lock:
            if not self.is_closed:
                await super().close()
                self.log.info('closed')
            else:
                self.log.warning('already closed')

    def __init__(
        self,
        url  : HTTP.URL,
        loop : AbstractEventLoop | None = None,
        ssl_context : SSLContext | None = None,
        **kw: Any,
    ):
        self.url = url = HTTP.URL.load(url).url
        self.transport = None

        self.kwargs = self._parse_parameters(kw or dict(url.query))
        self.kwargs.setdefault('context', ssl_context)

        self.loop = loop or get_event_loop()
        self.close_callbacks = CallbackCollection(self)

        self._closed = False
        self._close_called = False
        self._active: Event = Event()

    async def _on_connection_close(self, closing: Future):
        try:  # noqa: SIM105
            e = closing.exception()
        except CancelledError as e:  # noqa: F841
            ...
        self._active.clear()
        await self.close_callbacks(e)

    async def _on_connected(self):
        self._active.set()

    async def ready(self):
        await self._active.wait()

    @property
    def is_active(self):
        return self._active.is_set()

    def __bool__(self):
        return self.is_active


class Rabbit(dataclass.auto('config')):

    class Config(dataclass.flex):
        uri     : str
        options : dict     = field(default_factory=dict)
        timeout : Timeouts = field(default_factory=Timeouts)

        prefetch  : int = 0
        consistent: bool = True

    @classmethod
    def from_url(cls, url, /, **kw):
        return cls(config=cls.Config.load(kw | {'uri': url}))

    @Property.Cached
    def url(self):
        return HTTP.URL.load(self.config.uri)

    @Property.Cached
    async def connector(self):
        return await Connection.make(self.url, **self.config.options)

    @property
    async def connection(self):
        timeout = self.config.timeout
        connector = await self.connector
        head = f'{Who(connector, addr=True)}({connector})'

        async with connector.lock:
            while not connector:
                start = time()

                await retry(
                    connector.connect,
                    traceback = False,
                    intercept = exc.RabbitExceptions,
                    maximum   = timeout.max_delay,
                )()

                if connector:
                    delta = time() - start
                    if delta < 0.1:  # noqa: PLR2004
                        self.log.info(
                            f'{head} established fast as expected')
                    elif delta < 1.0:
                        self.log.notice(
                            f'{head} established after {delta:0.2f} sec')
                    else:
                        self.log.warning(
                            f'{head} established very slow: after {delta:0.2f} sec')
                    return connector

                await sleep(timeout.min_quant)

            self.log.debug(f'{head} already connected')
        return connector

    async def __aiter__(self):

        async def cleanup(channels):
            for channel in channels:
                with suppress(Exception):
                    await channel.close()
                    self.log.verbose(f'{channel=} closed')
            return []

        channels = []

        params_channel = {'publisher_confirms': self.config.consistent}
        params_prefetch = {'prefetch_count' : 1} if self.config.prefetch else {}

        while True:
            connection = await self.connection
            try:
                async with connection.channel(**params_channel) as channel:

                    channels.append(channel)

                    while True:
                        try:
                            qos = await channel.set_qos(**params_prefetch)
                            if not isinstance(qos, cmd.Basic.QosOk):
                                self.log.info(
                                    f'{channel=} {qos=} invalid state, reconnnect')
                                break

                        except rmq.ChannelInvalidStateError:
                            self.log.info(f'{channel=} invalid state, reconnnect')
                            break

                        if channel.is_initialized and channel.is_closed:
                            self.log.info(f'{channel=} closed, reconnect')
                            break

                        yield channel

            except CancelledError as e:
                self.log.warning(f'{connection=} cancelled: {e}')

            channels = await cleanup(channels)


class Routine(dataclass.auto('config')):

    class Config(dataclass.flex):
        name      : str
        key       : str
        exchange  : str

        limit     : int | None
        encoder   : str | None

        durable   : bool | None
        active    : bool = False
        autoack   : bool = False
        exclusive : bool = False
        temporary : bool = False

        # default settings
        exchange_type       : ExchangeType | str = ExchangeType.DIRECT
        exchange_persistent : bool = True

        # lock when nobody listening tasks queue, unlock with first worker
        wait_consumers      : bool | None

        # log all slow calls
        logging_threshold   : float = 1.0

        @classmethod
        def Apply(cls, func):  # noqa: N802

            @wraps(func)
            def wrapper(self, *args, **kw):

                from kalib import halt
                if args and issubclass(Is.classOf(args[0]), Routine.Config):
                    config, args = args[0], args[1:]
                else:
                    config = self.settings

                if kw:
                    config = config.copy(**kw)
                    kw = dict(config.extra_kwargs)

                return func(self, config, *args, **kw)

            return wrapper

        @classmethod
        def FilterNone(cls, func):  # noqa: N802
            @wraps(func)
            def wrapper(*args, **kw):
                return func(*args, **{k: v for k, v in kw.items() if v is not None})
            return wrapper

        @classmethod
        def Options(cls, *fields):  # noqa: N802
            idx = set(fields)

            def container(func):
                Options = dataclass.simple(f'{Who(func)}.Options', *fields)  # noqa: N806

                @wraps(func)
                def wrapper(*args, **kw):
                    options = dict(unique(kw, include=idx))
                    options = Options.load(options)
                    return func(
                        *args, options=options,
                        **dict(unique(kw, exclude=options.as_dict)))

                return wrapper
            return container

    class Carrier(dataclass.flex):

        config   : object
        channel  : object
        exchange : object
        queue    : object

        slots : int
        limit : int | None
        tasks : int | None
        clients: int | None


        @Property.Cached
        def str(self):
            msg = f'{self.exchange}'
            if self.queue:
                msg = f'{msg}.{self.queue}'
            return f'{msg}:{self.channel}'

    class Item(dataclass):
        body: bytes | str
        mime: str

        @Property.Cached
        def charset(self):
            return Str(self.body).charset

        @Property.Cached
        def hash(self):
            return digest(self.body)

    class Solution(dataclass):
        reconnect : bool = False
        sleep     : float | int = 0.0

        def __bool__(self):
            return self.reconnect

    #

    @Property.Cached
    def settings(self):  # shortcut, can be redirected in childs
        return self.config

    @Property.Cached
    def state(self):
        return Tag.new()

    @Property.Cached
    def timeout(self):
        return self.pool.config.timeout

    @Property.Cached
    @Config.Apply
    def durable(self, config):
        if config.durable is not None:
            return (
                (DeliveryMode.NOT_PERSISTENT, DeliveryMode.PERSISTENT)
                [bool(config.durable)])

    @Property.Cached
    async def carrier_lock(self):
        connector = await self.pool.connector
        return connector.lock

    # internal methods

    @classmethod
    def from_url(cls, url, /, **kw):
        config = cls.Config.load(kw)
        rabbit = Rabbit.from_url(url, **(config.as_dict | config.extra_kwargs))
        return cls(rabbit, config=config)

    def __init__(self, pool: Rabbit, /, **kw):
        super().__init__(**kw)
        self.pool = pool

    @Config.Apply
    def dumps(self, config, data) -> Item:
        if isinstance(data, dataclass):
            data = dict(data.as_dict)

        if config.encoder:
            body = dumps(data, config.encoder)
            mime = 'application/octet-stream'
        else:
            body = json.dumps(data, bytes=True)
            mime = 'application/json'

        return self.Item.load({'body': body, 'mime': mime})

    def loads(self, msg: Message):
        return loads(msg.body) or json.loads(msg.body)

    def make(self, item, **kw):
        if not isinstance(item, self.Item):
            item = self.dumps(item)

        kw.setdefault('headers', {})
        headers = kw['headers']
        headers.setdefault('hash', item.hash)

        if kw.get('meta'):
            headers.setdefault('meta', kw.pop('meta'))

        if kw.get('reply_to'):
            headers.setdefault('reply_to', kw['reply_to'])

        if (ttl := kw.pop('ttl', None)):
            kw.setdefault('expiration', ttl)

        kw.setdefault('body', item.body)
        kw.setdefault('content_type', item.mime)
        kw.setdefault('content_encoding', item.charset)

        if self.durable is not None:
            kw.setdefault('delivery_mode', self.durable)

        return Message(timestamp=time(), **kw)

    # rabbit entities wrappers

    @Config.Apply
    @Config.FilterNone
    @Config.Options('timeout')
    @log_long_wait
    async def exchange(self, config, channel, options, **kw):  # noqa: ARG002
        exchange = config.exchange
        try:
            return await channel.declare_exchange(
                name    = exchange,
                passive = not config.active,
                type    = config.exchange_type,
                durable = config.exchange_persistent,
                timeout = options.timeout)

        except rmq.ChannelNotFoundEntity as e:
            mode = ('passive', 'active')[bool(config.active)]
            msg = f'{exchange=} {mode=} not found'
            raise exc.ExchangeNotFoundError(config, msg) from e

    @Config.Apply
    @Config.FilterNone
    @Config.Options('timeout')
    @log_long_wait
    async def queue(self, config, channel, options, **kw):  # noqa: ARG002
        queue = config.name
        try:
            return await channel.declare_queue(
                name        = queue,
                passive     = not config.active,
                durable     = config.durable or None,
                exclusive   = config.exclusive or None,
                auto_delete = config.temporary or None,
                timeout = options.timeout)

        except rmq.ChannelNotFoundEntity as e:
            mode = ('passive', 'active')[bool(config.active)]
            msg = f'{queue=} {mode=} not found'
            raise exc.QueueNotFoundError(config, msg) from e

    @Config.FilterNone
    @Config.Options('timeout', 'routing_key')
    @log_long_wait
    async def bind(self, queue, exchange, options, **kw):

        result = await queue.bind(
            exchange    = exchange,
            timeout     = options.timeout,
            routing_key = options.routing_key,
        )

        if isinstance(result, cmd.Queue.BindOk):
            return True

        kw = json.repr(kw)
        self.log.error(
            f'{exchange=} --> {queue=} ({kw=}) something went wrong')

    async def done(self, queue, text, reset=False, **kw):
        self.log.verbose(text)
        result = await queue.cancel(consumer_tag=self.state.tag, **kw)
        if reset:
            self.state.reset()
        return result

    # carrier related methods

    @Config.Apply
    @Config.Options('check_limits', 'counters', 'timeout', 'routing_key')
    async def carrier(self, config, channel, options, **kw):  # noqa: ARG002

        timeout = options.timeout or self.timeout.max_delay
        consumers, limit, slots, tasks, queue = 0, config.limit or 0, 0, 0, None

        async with await self.carrier_lock:
            exchange = await self.exchange(config, channel, timeout=timeout)
            if config.name:
                queue = await self.queue(config, channel, timeout=timeout)
                await self.bind(
                    queue, exchange,
                    timeout     = timeout,
                    routing_key = options.routing_key,
                )

        if options.check_limits or options.counters:
            if config.name is None or not queue:
                msg = f"{queue=} ({config.name=}), couldn't get queue limits"
                self.log.fatal(msg)
                raise exc.ConfigurationError(config, msg)

            declaration = queue.declaration_result
            consumers = declaration.consumer_count

            if not consumers and options.check_limits:
                msg = f'no consumers on {queue.name=}'
                self.log.warning(msg)
                if config.wait_consumers:
                    raise exc.NoActiveConsumersError(config, msg)

            tasks = declaration.message_count
            slots = max(0, limit - tasks)

            if not slots and limit > 0 and options.check_limits:
                msg = (
                    f'no free slots in {queue.name=}: awaiting {tasks=} / '
                    f'queue {limit=}, free {slots=}')
                self.log.info(msg)
                raise exc.NotAvialableQueueSlotsError(config, msg)

        return self.Carrier.load({
            'config'  : config,
            'channel' : channel,
            'exchange': exchange,
            'queue'   : queue,

            'limit'   : limit,
            'slots'   : slots,
            'tasks'   : tasks,
            'clients' : consumers,
        })

    @Property.Cached
    def _carrier_and_exceptions(self):
        return partial(
            retry,
            self.carrier,
            passthru  = True,
            traceback = False,
            limit     = self.timeout.reconnect,
            maximum   = self.timeout.max_delay), {
                TimeoutExceptions: True, (
                exc.ExchangeNotFoundError,
                exc.QueueNotFoundError,
                rmq.ChannelPreconditionFailed,
            ): False}

    @property
    def _consume_carrier(self):
        func, exceptions = self._carrier_and_exceptions
        return func(
            intercept = exceptions | {rmq.DuplicateConsumerTag: True})

    @property
    def _publish_carrier(self):
        func, exceptions = self._carrier_and_exceptions
        return func(
            intercept = exceptions | {exc.ConsumerNotReadyError: True})

    # main logic

    @Config.Apply
    def solution(self, config, exception):
        header  = f'catch ({Who(exception)}) {exception=}'
        expected, level, sleep = True, Logging.Error, self.timeout.min_delay

        kw = {}
        if issubclass(exception, TimeoutExceptions):
            level, sleep = Logging.Verbose, self.timeout.min_quant
            message = (
                f'{header}: {config.exchange}.{config.name}.{config.key} hit timeout')

        elif issubclass(exception, rmq.ChannelPreconditionFailed):
            expected = False
            message = f'{header}: configuration unexpected exception'

        elif issubclass(exception, rmq.ChannelInvalidStateError):
            level, sleep = Logging.Verbose, self.timeout.min_quant
            message = f'{header}: just refresh connection'

        elif issubclass(exception, (pam.PAMQPException, rmq.AMQPError)):
            message = f'{header}: generic rabbit unexpected exception'

        elif issubclass(exception, exc.EntityNotFoundError):
            kw['once'] = True
            expected = False
            message = f"{header}: bus entities isn't created"

        else:
            expected = False
            message = f'{header}: unexpected exception'

        trace = not expected or level is Logging.Error
        self.log.by_level(level)(message, trace=trace, stack=1, shift=-1, **kw)
        return self.Solution.load(expected, sleep)

    @Config.Apply
    @Config.Options('check_limits', 'routing_key', 'timeout')
    async def send(  # noqa: PLR0912, PLR0915, PLR0913, C901
        self, config, channel, something, /,
        message_builder=None, options=None, **kw,  # noqa: ARG002
    ):
        repeatable, result = False, []
        if message_builder and not Is.callable(message_builder):
            msg = f"({Who(message_builder)}) {message_builder=} isn't callable"
            raise TypeError(msg)

        consistent = self.pool.config.consistent
        params = {k: v for k, v in options.as_dict.items() if v is not None}
        check_limits = params.pop('check_limits', False)
        params = {
            'routing_key' : config.key,
            'timeout'     : self.timeout.max_delay,
        } | params

        if isinstance(something, Message):
            tasks = (something,)
        elif isinstance(something, list | set | tuple):
            tasks = tuple(something)
        else:
            repeatable, tasks = True, []

        while True:
            try:
                if channel.is_initialized and channel.is_closed:
                    msg = f'{Who.Is(channel)} closed'
                    raise exc.ChannelClosedError(config, msg)  # noqa: TRY301

                carrier = await self._publish_carrier(
                    config, channel, check_limits=check_limits, **params)

                if repeatable and not tasks:
                    tasks.extend(await something(carrier))

                amount = (config.limit or 0) and carrier.slots or len(tasks)
                for task in tuple(tasks[:amount]):
                    if isinstance(task, Message):
                        msg = task

                    elif message_builder:
                        msg = await message_builder(task)

                    if not isinstance(msg, Message):
                        msg = (
                            f"({carrier.str}) task isn't aio_pika.Message, "
                            f"({Who(msg)}) {msg=}")

                        await sleep(self.timeout.min_quant)
                        raise exc.ConfigurationError(carrier, msg)  # noqa: TRY301

                    start = time()
                    response = await carrier.exchange.publish(msg, **params)
                    spent = time() - start

                    maybe = '' if consistent else 'maybe '
                    msg = f'({carrier.str}) message ({len(msg.body)} bytes) {maybe}ok'

                    if consistent and (
                        isinstance(response, DeliveredMessage)
                        or not isinstance(response, cmd.Basic.Ack)
                    ):
                        msg = f"response isn't acknowledged: {response}"
                        if carrier.queue:
                            msg = (
                                f"{msg}, looks like consumer queue {config.name} "
                                f"hasn't bounded (or created) to {config.exchange=}")
                        else:
                            msg = (
                                f"{msg}, looks like {config.exchange=} isn't created "
                                f"or no queues found")

                        if config.active and isinstance(response, DeliveredMessage):
                            error = exc.PublishingMessageUnrecoverableError
                        else:
                            error = exc.PublishingMessageError

                        await sleep(self.timeout.min_quant)
                        raise error(carrier, msg)  # noqa: TRY301

                    elif not consistent:  # noqa: RET506
                        self.log.debug(
                            f'publisher confirmation disabled by {consistent=}')

                    if spent > config.logging_threshold:
                        self.log.warning(f'{msg}, but too slow, {spent:0.2f} sec')
                    else:
                        self.log.debug(msg)

                    tasks = tasks[1:]

                    if hasattr(task, 'mark_published'):
                        await task.mark_published()
                    elif hasattr(self, 'mark_published'):
                        await self.mark_published(task)

                    result.append(response)

                if not repeatable and not tasks:
                    return tuple(response)

                await sleep(self.timeout.min_quant)
                continue

            except exc.PublishingMessageError as e:
                msg = (
                    f'catch {Who(e)}, looks like broken or '
                    f'misconfigured connection: {e}')
                self.log.error(msg)  # noqa: TRY400
                break

            except exc.ConsumerNotReadyError:
                try:
                    addr = carrier.str
                except UnboundLocalError:
                    addr = config.name

                msg = (
                    f'({addr}) nobody listening or not free slots found '
                    f'for {self.timeout.reconnect:0.2f} sec, wait again')
                self.log.verbose(msg)

                await sleep(self.timeout.min_quant)
                if config.wait_consumers:
                    continue
                raise

            except Exception as e:
                if reconnect := self.solution(e):
                    await sleep(reconnect.sleep)
                    break
                await sleep(self.timeout.min_quant)
                raise

    @Config.Apply
    @Config.Options('timeout')
    async def recv(self, config, channel, callback, options, **kw):  # noqa: ARG002
        consumer, timeout = self.state, options.timeout or self.timeout.max_delay
        while True:
            if channel.is_initialized and channel.is_closed:
                msg = f'{Who.Is(channel)} closed'
                raise exc.ChannelClosedError(config, msg)

            carrier = await self._consume_carrier(config, channel, timeout=timeout)
            discard = partial(self.done, carrier.queue, timeout=timeout)

            if consumer:
                await discard(f'({carrier.str}) {consumer.tag=} restart')
            else:
                self.log.verbose(f'({carrier.str}) declare {consumer.tag=}')

            settings = {}
            if config.autoack is True:
                settings['no_ack'] = config.autoack

            try:
                async with carrier.queue.iterator(
                    consumer_tag = consumer.tag,
                    timeout      = timeout,
                    **settings,
                ) as inbox:

                    async for task in inbox:
                        result = await callback(task)
                        if result:
                            response = json.repr(result)
                            function = '' if Is.Internal(callback) else (
                                f'({Who(callback)}) ')

                            await discard(
                                f'({carrier.str}) {consumer.tag=}; got '
                                f'{function}{response=}', reset=True)
                            return result

            except TimeoutExceptions:
                msg = (
                    f'({carrier.str}) nothing received after '
                    f'{timeout:0.2f} sec, reloop and wait again')
                self.log.verbose(msg)

            except rmq.DuplicateConsumerTag as e:
                self.state.reset()
                self.log.error(  # noqa: TRY400
                    f'({carrier.str}) catch {Who(e)}, recreate consumer tag')

            except Exception as e:
                if reconnect := self.solution(e):
                    await sleep(reconnect.sleep)
                    return
                await sleep(self.timeout.min_quant)
                raise

            await sleep(self.timeout.min_quant)

    @Config.Apply
    async def perform(self, config, method, callback, *args, **kw):
        async for channel in self.pool:
            try:
                result = await method(config, channel, callback, *args, **kw)

            except exc.ChannelClosedError:
                continue

            except (exc.ExchangeNotFoundError, exc.QueueNotFoundError) as e:
                msg = f'catch {Who(e)}, mode {config.active=}'

                if not config.active:
                    self.log.warning(
                        f'{msg}, wait {self.timeout.min_delay:0.2f}s and try again')
                    await sleep(self.timeout.min_delay)
                    continue

                self.log.fatal(f'{msg}, declaration unexpected exception')
                raise

            if result is not None:
                return result

    # high level interface

    message_maker = make

    @Config.Apply
    @attach('carrier', _publish_carrier)
    async def publish(self, config, callback, /, **kw):
        kw.setdefault('check_limits', True)
        await self.perform(config, self.send, callback, **kw)

    @Config.Apply
    @attach('carrier', _consume_carrier)
    async def consume(self, config, callback, /, **kw):
        await self.perform(config, self.recv, callback, **kw)



import asyncio  # noqa: I001
import builtins
from abc import ABC
from collections import defaultdict, deque
from collections.abc import Iterable
from functools import partial, reduce, wraps
from itertools import combinations
from logging import Logger
from math import exp
from operator import or_
from random import uniform
from time import time
from typing import Any, ClassVar

from kalib import Is, Logging, Property, Who, exception, sort, unique, dataclass


def incremental_delay(limit=None, deviation=None, function=exp):
    def wrapper(attempt, *args, **kw):
        result = function(attempt, *args, **kw)

        if deviation:
            bias = uniform(1.0 - deviation, 1.0 + deviation)  # noqa: S311
            result = uniform(result * bias, result / bias)  # ~5.54..9.85  # noqa: S311
        return float(limit if limit and result >= limit else result)
    return wrapper


def to_tuple(x):
    return tuple(x or ()) if Is.iterable(x) else (x,)


def make_properties(klass):

    def make_wrapper(name):
        def wrapper(cls):
            return to_tuple(getattr(cls, name))
        return wrapper

    for name in klass.attrs:
        func = name.lower()
        wrapper = make_wrapper(name)
        wrapper.__name__ = func
        wrapper.__qualname__ = f'{Who(klass, full=False)}.{func}'

        if hasattr(klass, func):
            msg = f'property {wrapper.__qualname__} already exists'
            raise ValueError(msg)
        setattr(klass, func, Property.Class(wrapper))

    return klass


def positive_numeric(func):
    @wraps(func)
    def wrapped(self, *args, **kw):
        result = func(self, *args, **kw)
        if not isinstance(result, float | int) or result < 0:
            msg = (
                f'{Who(func)} must be positive numeric, '
                f'not ({Who(result)}) {result=}')
            raise ValueError(msg)
        return result
    return wrapped


class Exceptions(dataclass):

    recover : list[Exception] = ()
    expect  : list[Exception] = ()
    suppress: list[Exception] = ()


@make_properties
class ExceptionsContainer(Logging.Mixin):

    attrs = ('Expected', 'Recoverable')

    @Property.Class.Cached
    def order(cls):
        result = {}
        attrs = set(cls.attrs)
        keys = tuple(cls.__dict__)

        if attrs & set(keys):
            for key in keys:
                if key in attrs:
                    exceptions = getattr(cls, key.lower())
                    if exceptions:
                        result[exceptions] = key == 'Recoverable'

        return dict(result)


class VerboseError(ABC, Exception):
    def __init__(self, *args):
        if len(args) == 2:  # noqa: PLR2004
            item, message = args
            super().__init__(f'({Who(item)}) {item!r}: {message}')

        elif len(args) == 1:
            super().__init__(args[0])

        else:
            msg = (
                f'{Who(self)} can use 1 or 2 arguments, '
                f'not {len(args)}: {args!r}')
            raise ValueError(msg)


class RetryError(Exception):
    ...


class TaskTimedoutError(RetryError):
    ...


class WrappedTaskTimeoutError(RetryError):
    ...


TimeoutExceptions = (
    TimeoutError,
    asyncio.TimeoutError,
    builtins.TimeoutError,
    TaskTimedoutError)


class NetworkExceptions(ExceptionsContainer):
    Recoverable = (
        ConnectionError,
        ConnectionRefusedError,
        *TimeoutExceptions)


class BuiltinExceptions(ExceptionsContainer):
    Expected = (
        AttributeError, KeyError, ValueError,
        AssertionError, BufferError, EOFError, ImportError, LookupError,
        MemoryError, NameError, ReferenceError, StopAsyncIteration,
        StopIteration, SyntaxError, SystemError, TypeError, ValueError)



class Retry(dataclass.auto('config')):
    ContainerMixin = ExceptionsContainer
    _exceptions_groups_cache: ClassVar[dict] = {}

    class Config(dataclass.flex):

        # global limits
        count : int = 0  # max tries count
        limit : bool | int | float = 0.0  # relative +sec or absolute unix timestamp

        # per task limits
        timeout : float = 0.0  # max function call execution time
        minimum : float = 0.0  # min sleep time between retries
        maximum : float = 0.0  # max sleep time

        # basic logging configuration
        passthru  : bool = False # passthru exception when deadline reached
        traceback : bool | None = None  # show traceback always or only first time

        # exceptions configuration
        logging   : dict[Exception, Any] | bool | None = False
        intercept : (
            Exception |
            Iterable[Exception] |
            dict[Exception, bool] |
            ExceptionsContainer |
            Any)

    @Property.Cached
    def logger(self):
        return self._logger or Logging.get(self)

    def __init__(self, func, logger: Logger | None = None, /, **config):
        if logger is not None and not Is.subclass(logger, Logger):
            msg = (
                f'logger must be instance of {Who(Logger)}, '
                f'not ({Who(logger)}) {logger=}')
            raise TypeError(msg)

        # static data
        self._func = func
        self._logger = logger

        # internal variables
        self._start = None
        self._counter = 0
        self._history = []
        self._timeouts = deque(maxlen=2 ** 10)

        super().__init__(config=config)

    # static properties

    @Property.Cached
    def function(self):
        return self._func

    @Property.Cached
    @positive_numeric
    def timeout(self):
        timeout = self.config.timeout
        return (
            float(timeout)
            if isinstance(timeout, float | int) and timeout > 0 else 0.0)

    @Property.Cached
    def logging(self):
        logging = self.config.logging

        if logging is None or isinstance(logging, bool):
            return {Exception: logging}

        if not isinstance(logging, dict):
            msg = f'({Who(logging)}) {logging=} must me dict or bool'
            raise TypeError(msg)

        return self.solutions(logging)

    @property
    def show_traceback(self):
        traceback = self.config.traceback
        return traceback or (traceback is None and self.counter == 1)

    @Property.Cached
    @positive_numeric
    def minimum(self):
        return self.config.minimum

    @Property.Cached
    @positive_numeric
    def maximum(self):
        return self.config.maximum

    @Property.Cached
    @positive_numeric
    def count(self):
        return self.config.count

    @property
    @positive_numeric
    def deadline(self):
        limit = self.config.limit

        if limit is True:
            if not self.maximum:
                self.log.warning(
                    f'{limit=}, but {self.maximum=}, disable deadlining')
            limit = self.maximum

        elif not isinstance(limit, float | int):
            return limit

        if not limit:
            return 0.0

        if limit <= 365 * 10:
            limit += time()

        return limit

    @Property.Cached
    def sleep_time(self):
        return incremental_delay(self.maximum)

    # active properties

    @property
    def counter(self):
        return self._counter

    @counter.setter
    def counter(self, counter):
        if not isinstance(counter, int):
            msg = f'({Who(counter)}) {counter=} must me positive integer'
            raise TypeError(msg)
        counter = int(counter)
        self._counter = counter
        return counter

    @property
    def exception(self):
        try:
            return self._history[-1]

        except IndexError:
            msg = f'{Who(self)}.exception is never set'
            raise ValueError(msg)  # noqa: B904, TRY200

    @exception.setter
    def exception(self, e):
        if not Is.subclass(e, Exception):
            msg = f'({Who(e)}) {e=} must be inherited from Exception'
            raise TypeError(msg)

        self._history.append(e)
        return self._history[-1]

    @property
    def start(self):
        start = self._start
        if start is None:
            msg = f'({Who(start)}) {start=} is never set'
            raise ValueError(msg)
        return start

    @start.setter
    def start(self, start):
        if not isinstance(start, float | int):
            msg = f'({Who(start)}) {start=} must me positive float'
            raise TypeError(msg)
        start = float(start)

        self._start = start
        return start

    @property
    def average(self):
        if not self._timeouts:
            msg = f'{Who(self)}.average never set'
            raise ValueError(msg)
        return sum(self._timeouts) / len(self._timeouts)

    @average.setter
    def average(self, start):
        if not isinstance(start, float | int):
            msg = f'({Who(start)}) {start=} must me positive float'
            raise TypeError(msg)

        value = time() - float(start)
        self._timeouts.append(value)
        return value

    # exceptions order configuration methods

    def validate(self, order):
        seen = set()
        shadows = []
        for no, stage in enumerate(order):
            if not no:
                seen.update(stage)
                continue

            mode = ('expected', 'recoverable')[order[stage]]
            for e in stage:
                if Is.subclass(e, to_tuple(seen)):
                    for num, ee in enumerate(order):
                        if Is.subclass(e, to_tuple(ee)):
                            shadows.append(
                                f'{mode} {Who(e)} at stage '
                                f'{no} shadowed by stage {num}')
                            break
            seen.update(stage)

        if shadows:
            self.log.trace(
                f'some exceptions never intercept: {"; ".join(shadows)}')

        return dict(order)

    def flattened_solutions(self, order):  # noqa: PLR0912

        def by_depth(x):
            return str(Who(x).count('.')) + Who(x)

        last = None
        result = []
        for stage, solution in order.items():
            if last is None:
                result.append( [solution, *stage] )
                last = solution
                continue

            if solution is not last:
                result.append([solution])
                last = solution

            result[-1].extend(stage)

        union_stages = {}
        for solution, *stage in result:
            union_stages[tuple(stage)] = solution

        ordered_stages = []
        for stage in reversed(union_stages):
            try:
                ordered_stages.append(self._exceptions_groups_cache[stage])

            except KeyError:
                classes_to_remove = set()
                parent_to_childs = defaultdict(set)
                childs_to_parent = defaultdict(set)

                for foo, bar in combinations(stage, 2):
                    if foo is bar:
                        continue

                    if Is.subclass(foo, bar):
                        parent_to_childs[bar].add(foo)
                        childs_to_parent[foo].add(bar)

                    elif Is.subclass(bar, foo):
                        parent_to_childs[foo].add(bar)
                        childs_to_parent[bar].add(foo)

                parent_to_childs = dict(parent_to_childs)
                childs_to_parent = dict(childs_to_parent)

                if parent_to_childs:
                    duplicated = set(unique(reduce(or_,
                        ((parent_to_childs.get(e, set()) | {e})
                            for e in childs_to_parent))))

                    for e in duplicated:
                        classes_to_remove.add(e)
                        if e in parent_to_childs:
                            classes_to_remove.update(parent_to_childs[e])

                    pivots = defaultdict(set)
                    for e in classes_to_remove:
                        for pivot in childs_to_parent[e] - classes_to_remove:
                            pivots[pivot].add(e)

                    result = []
                    for e in sort(pivots, key=by_depth):
                        result.append(
                            f'({", ".join(sort(map(Who, pivots[e])))}) '
                            f'-> {Who(e)}')

                    self.log.debug(
                        f'remove duplicated: {"; ".join(result)}')

                reordered = tuple(sorted(
                    unique(stage, exclude=classes_to_remove),
                    key=lambda x: len(Is.classOf(x).__mro__), reverse=True))

                ordered_stages.append(reordered)
                self._exceptions_groups_cache[stage] = reordered

        return self.validate(
            dict(zip(
                reversed(ordered_stages),
                union_stages.values(), strict=True)))

    def solutions(self, something):
        invalid_message = (
            f'intercepted exceptions order can be Exception, '
            f'dict[Exception, bool], Iter[Exception], '
            f'or {Who.Is(ExceptionsContainer)} '
            f'not {Who.Is(something)}; '
            f'check closest {Who(self)} invocation')

        def make_tuple(x):
            if Is.subclass(x, Exception):
                return {(x,): True}

            elif x is BaseException:
                msg = f'never try to catch BaseException: {Who.Is(x)}'
                raise ValueError(msg)

            elif Is.subclass(x, ExceptionsContainer):
                return make_tuple(x.order)

            elif Is.subclass(x, asyncio.exceptions.CancelledError):
                return {(x,): True}

            result = {}

            def propagate_order(result, order):
                result = dict(result)
                for k, v in order.items():
                    if k not in result:
                        result[to_tuple(k)] = v
                    else:
                        pass
                return dict(result)

            if isinstance(x, dict):
                return propagate_order(result, x)

            elif isinstance(x, list | set | tuple):
                for i in x:
                    result = propagate_order(result, make_tuple(i))
                return result

            raise TypeError(invalid_message)

        if something is all:
            something = [{Exception: True}]

        elif something is any:
            something = [{BuiltinExceptions.Expected: False, Exception: True}]

        elif not something:
            raise ValueError(invalid_message)

        order = make_tuple(something)
        if not order:
            raise ValueError(invalid_message)

        exceptions = self.flattened_solutions(order)
        if (
            len(exceptions) == 1 and
            not (key := next(iter(exceptions))) and
            isinstance(key, tuple)
        ):
            self.log.trace(
                f"empty recoverable {exceptions=} order, let's intercept "
                f'any exception! (but rewrite with intercept=any kwarg)')
            exceptions = {(Exception,): True}

        return dict(exceptions)

    def get_exception_value_from_order(self, exceptions, exception, /, **kw):
        for stage, value in exceptions.items():
            if Is.subclass(exception, stage):
                return value
        try:
            return kw['default']
        except KeyError:
            msg = f'{exception=} not in {exceptions=}'
            raise KeyError(msg)  # noqa: B904, TRY200

    def intercept(self, exception, message, full=False, level=None, **kw):
        if full:
            message = f'{Who(exception)}: {exception=}; {message}'

        level = level or self.get_exception_value_from_order(
            self.logging, exception, default=Logging.Verbose)

        kw.setdefault('trace', self.show_traceback)
        self.log.by_level(level)(message, **kw)

    def passthru(self, message):
        exception = self.exception

        message = f'{message}, raise ({Who(exception)})'
        if not Is.Class(exception):
            message = f'{message} {exception=}'
        self.intercept(exception, message)


        if (
            not self.config.passthru or
            Is.classOf(exception) is WrappedTaskTimeoutError
        ):
            raise TaskTimedoutError from exception
        raise exception

    # prepared exceptions properties

    @Property.Cached
    def expected(self):
        exceptions = []
        for e, recoverable in self.exceptions.items():
            if recoverable:
                exceptions.extend(e)
        return next(iter(self.solutions(exceptions)))

    @Property.Cached
    def exceptions(self):
        return self.solutions(self.config.intercept)

    # main methods

    def solution(self, message, exceptions, deadline, start):
        now = time()

        self.average = start
        exception = self.exception

        if Is.subclass(exception, WrappedTaskTimeoutError):
            can_recover = True
        else:
            try:
                can_recover = self.get_exception_value_from_order(exceptions, exception)
            except KeyError:
                return

        self.counter += 1
        count = self.count

        message = (
            f'{self.counter}/{count}. ' if count else f'{self.counter}. '
            f'after {now - start:0.3f}/{now - self.start:0.3f} sec, {message}')

        if count and self.counter >= count:
            message = f'{message}, maximum {count=} reached'
            self.passthru(message)

        if not can_recover:
            self.intercept(
                exception, f'{message} got expected {Who(exception)}, raise')
            raise exception

        wait = self.sleep_to
        message = f'{message} got {Who(exception)}'

        if deadline:
            if now > deadline:
                overdue = now - deadline
                message = (
                    f'{message}, deadline +{(now - self.start):0.2f} '
                    f'sec reached')
                if overdue >= 0.1:  # noqa: PLR2004
                    message = '{message}, overdue {overdue:0.2f} sec'
                self.passthru(message)

            plan = deadline - now - self.average
            if plan < 0:
                message = (
                    f'{message}, deadline {deadline:0.2f} '
                    f'(average {self.average:0.2f}) after '
                    f'{(now - deadline):0.2f} sec, can be reached')
                self.passthru(message)

            if plan < wait:
                message = (
                    f'{message}, retry wait {wait:0.2f} > deadline '
                    f'{(deadline - now):0.2f} - {self.average:0.2f}, '
                    f'shrink sleep time to {plan:0.2f} sec')
                wait = plan

        message = f'{message}, sleep {wait:0.2f} sec'
        if deadline and (delta := (deadline - time())) and delta < wait:
            message = f'{message}, last chance after {delta:0.2f} sec'

        self.intercept(exception, message)
        return wait

    @property
    def sleep_to(self):
        return max(self.minimum, self.sleep_time(self.counter))

    @classmethod
    def make(cls, *args, **kw):
        self = cls(*args, **kw)

        def override_exception(data, key, value):
            if (order := data.pop(key, None)):
                if Is.subclass(order, Exception):
                    return {(order,): value}

                if not isinstance(order, list | set | tuple):
                    msg = (
                        f'{Who(order)} {order=} must be '
                        f'Exception, tuple or list')
                    raise TypeError(msg)

                return {to_tuple(order): value}

            return {}

        @wraps(self.function)
        async def retry(*args, **kw):  # noqa: PLR0912
            self.start = time()

            order = (
                override_exception(kw, 'recover', False),  # noqa: FBT003
                override_exception(kw, 'expect', True),  # noqa: FBT003
                self.exceptions)

            kw = {**self.config.extra_kwargs, **kw}
            header = f'{Who(self.function)}({args=}, {kw=})'
            deadline = self.deadline

            exceptions = {}
            for stage in order:
                for key, value in stage.items():
                    if key in exceptions:
                        msg = f'{header} {key=} already defined, override with {value=}'
                        self.log.warning(msg)
                    exceptions[key] = value
            exceptions = self.solutions(exceptions)

            order = '; '.join(
                f'{("throw expected", "handle recoverable")[recoverable]} '
                f'[{", ".join(map(Who, stage))}]'
                    for stage, recoverable in exceptions.items())
            self.log.verbose(f'{header} {order}', once=True)

            solution = partial(self.solution, header, exceptions, deadline)
            while True:
                task = self.function(*args, **kw)
                coro = asyncio.create_task(task)

                start = time()
                try:
                    if self.timeout:
                        return await asyncio.wait_for(coro, timeout=self.timeout)
                    else:
                        return await coro

                except Exception as exception:

                    if Is.substance(exception, asyncio.TimeoutError):
                        state = coro._state  # noqa: SLF001

                        if state == 'CANCELLED':
                            exception, timeout = WrappedTaskTimeoutError, self.timeout
                            self.intercept(
                                exception, f'{header} catch wrapper '
                                f'{timeout=}, {Who.Is(state)=}')

                        elif state == 'FINISHED':
                            self.intercept(
                                exception, f'{header} throw '
                                f'{exception=}, {Who.Is(state)=}')
                        else:
                            msg = f'unexpected task {task=} ({coro=}) {state=}'
                            raise ValueError(msg)  # noqa: B904, TRY200

                    self.exception = exception
                    sleep = solution(start)

                    if sleep is not None:
                        await asyncio.sleep(sleep)
                        continue

                    self.intercept(
                        exception, f'{header} got unhandled '
                        f'{Who(exception)}, after {self.counter} attempt',
                        level=Logging.Error)
                    raise

        retry.expectations = self.expected
        retry.cancelled = TaskTimedoutError
        return retry


def format_exception(e):
    return exception(e).as_dict | {
        'class'  : Who(e),
        'source' : Who.File(e)}


retry = Retry.make



from functools import wraps
from inspect import iscoroutinefunction
from time import time

from kalib import Who, json


def attach(name, method):
    def wrapper(func):
        setattr(func, name, method.fget)
        return func
    return wrapper


def log_long_wait(func):

    def get_threshold(node):
        config = getattr(node, 'settings', node.config)
        return getattr(config, 'logging_threshold', 1.0)

    if iscoroutinefunction(func):
        @wraps(func)
        async def wrapper(node, *args, **kw):
            start = time()
            result = await func(node, *args, **kw)
            spent = time() - start


            if spent >= get_threshold(node):
                args, kw = json.repr((node, *args)), json.repr(kw)
                msg = f'long {spent:0.2f}s call {Who(func)}(*{args=}, **{kw=})'
                node.log.warning(msg, stack=1)

            return result

    else:
        @wraps(func)
        def wrapper(node, *args, **kw):
            start = time()
            result = func(node, *args, **kw)

            spent = time() - start
            if spent >= get_threshold(node):
                args, kw = json.repr((node, *args)), json.repr(kw)
                msg = f'long {spent:0.2f}s call {Who(func)}(*{args=}, **{kw=})'
                node.log.warning(msg, stack=1)

            return result

    return wrapper



import asyncio.exceptions as aio

import aiormq.exceptions as rmq
import pamqp.exceptions as pam

from karq.retry import ExceptionsContainer, NetworkExceptions, VerboseError


class RabbitError(VerboseError):
    ...


class ConfigurationError(VerboseError):
    ...


class EntityNotFoundError(ConfigurationError):
    ...


class ExchangeNotFoundError(EntityNotFoundError):
    ...


class QueueNotFoundError(EntityNotFoundError):
    ...


class MessagesFlowControrError(VerboseError):
    ...


class ChannelClosedError(MessagesFlowControrError):
    ...


class MessagesFlowError(VerboseError):
    ...


class ConsumerNotReadyError(MessagesFlowError):
    ...


class NoActiveConsumersError(ConsumerNotReadyError):
    ...


class NotAvialableQueueSlotsError(ConsumerNotReadyError):
    ...


class ProducingFlowError(MessagesFlowError):
    ...


class PublishingMessageUnrecoverableError(MessagesFlowError):
    ...


class PublishingMessageError(MessagesFlowError):
    ...


class ProduceTasksAmountError(MessagesFlowError):
    ...


class IncorrectProducedTasksAmountError(ProduceTasksAmountError):
    ...


class IncorrectToProduceTasksAmountError(ProduceTasksAmountError):
    ...


class RabbitExceptions(ExceptionsContainer):

    Recoverable = (
        aio.CancelledError,
        pam.AMQPChannelError,
        pam.AMQPConnectionForced,
        rmq.AMQPConnectionError,
        rmq.ChannelClosed,
        rmq.ChannelInvalidStateError,
        rmq.ConnectionClosed,
        *NetworkExceptions.Recoverable)

    Expected = (
        pam.PAMQPException,
        rmq.AMQPError,
    )

    NotFound = (
        rmq.ChannelNotFoundEntity,
        EntityNotFoundError,
    )
